import json
import os
import signal
import sys
from zipfile import BadZipfile
from zlib import error as zlib_error

from defusedxml.common import DefusedXmlException

import validator
from validator import decorator
from validator.chromemanifest import ChromeManifest
from validator.json_parser import ManifestJsonParser
from validator.opensearch import detect_opensearch
from validator.rdf import RDFException, RDFParser
from validator.typedetection import detect_type
from validator.xpi import XPIManager

from constants import (PACKAGE_ANY, PACKAGE_EXTENSION, PACKAGE_SEARCHPROV,
                       PACKAGE_THEME)


types = {0: 'Unknown',
         1: 'Extension/Multi-Extension',
         2: 'Full Theme',
         3: 'Dictionary',
         4: 'Language Pack',
         5: 'Search Provider'}

assumed_extensions = {'jar': PACKAGE_THEME,
                      'xml': PACKAGE_SEARCHPROV}


def prepare_package(err, path, expectation=0, for_appversions=None,
                    timeout=-1):
    """Prepares a file-based package for validation.

    timeout is the number of seconds before validation is aborted.
    If timeout is -1 then no timeout checking code will run.
    """

    package = None
    try:
        # Test that the package actually exists. I consider this Tier 0
        # since we may not even be dealing with a real file.
        if not os.path.isfile(path):
            err.error(('main', 'prepare_package', 'not_found'),
                      'The package could not be found')
            return

        # Pop the package extension.
        package_extension = os.path.splitext(path)[1]
        package_extension = package_extension.lower()

        def timeout_handler(signum, frame):
            raise validator.ValidationTimeout(timeout)

        if timeout != -1:
            signal.signal(signal.SIGALRM, timeout_handler)
            signal.setitimer(signal.ITIMER_REAL, timeout)

        if package_extension == '.xml':
            test_search(err, path, expectation)
        elif package_extension not in ('.xpi', '.jar'):
            err.error(('main', 'prepare_package', 'unrecognized'),
                      'The package is not of a recognized type.')
        else:
            package = open(path, 'rb')
            test_package(err, package, path, expectation, for_appversions)

    except validator.ValidationTimeout:
        err.system_error(
            msg_id='validation_timeout',
            message='Validation has timed out',
            signing_severity='high',
            description=('Validation was unable to complete in the allotted '
                         'time. This is most likely due to the size or '
                         'complexity of your add-on.',
                         'This timeout has been logged, but please consider '
                         'filing an issue report here: '
                         'https://bit.ly/1paBbSy'),
            exc_info=sys.exc_info())

    except Exception:
        err.system_error(exc_info=sys.exc_info())

    finally:
        # Remove timers and signal handlers regardless of whether
        # we've completed successfully or the timer has fired.
        if timeout != -1:
            signal.setitimer(signal.ITIMER_REAL, 0)
            signal.signal(signal.SIGALRM, signal.SIG_DFL)

        if package:
            package.close()

        decorator.cleanup()


def test_search(err, package, expectation=0):
    'Tests the package to see if it is a search provider.'

    expected_search_provider = expectation in (PACKAGE_ANY,
                                               PACKAGE_SEARCHPROV)

    # If we're not expecting a search provider, warn the user and stop
    # testing it like a search provider.
    if not expected_search_provider:
        return err.warning(('main',
                            'test_search',
                            'extension'),
                           'Unexpected file extension.')

    # Is this a search provider?
    detect_opensearch(err, package, listed=err.get_resource('listed'))

    if expected_search_provider and not err.failed():
        err.detected_type = PACKAGE_SEARCHPROV


def test_package(err, file_, name, expectation=PACKAGE_ANY,
                 for_appversions=None):
    'Begins tests for the package.'

    # Load up a new instance of an XPI.
    try:
        package = XPIManager(file_, mode='r', name=name)

        has_package_json = 'package.json' in package
        has_manifest_json = 'manifest.json' in package
        has_install_rdf = 'install.rdf' in package

        # install.rdf? | package.json? | manifest.json? | error | use-file
        # Yes          | No            | No             | No    | install.rdf
        # Yes          | Yes           | No             | No    | install.rdf
        # Yes          | No            | Yes            | No    | install.rdf
        # No           | No            | Yes            | No    | manifest.json
        # No           | No            | No             | Yes   | install.rdf
        # No           | Yes           | No             | No    | package.json
        # No           | No            | Yes            | Yes   | install.rdf
        if has_package_json:
            _load_package_json(err, package, expectation)
        if has_manifest_json:
            _load_manifest_json(err, package, expectation)
        if has_install_rdf:
            _load_install_rdf(err, package, expectation)
    except IOError:
        # Die on this one because the file won't open.
        err.error(('main', 'test_package', 'unopenable'),
                  'The XPI could not be opened.')
        return
    except (BadZipfile, zlib_error):
        # Die if the zip file is corrupt.
        err.error(('submain', '_load_install_rdf', 'badzipfile'),
                  error='Corrupt ZIP file',
                  description='We were unable to decompress the zip file.')
        return

    if package.extension in assumed_extensions:
        assumed_type = assumed_extensions[package.extension]
        # Is the user expecting a different package type?
        if expectation not in (PACKAGE_ANY, assumed_type):
            err.error(('main', 'test_package', 'unexpected_type'),
                      'Unexpected package type (found theme)')

    test_inner_package(err, package, for_appversions)


def _load_install_rdf(err, package, expectation):
    try:
        install_rdf = RDFParser(err, package.read('install.rdf'))
    except (RDFException, DefusedXmlException) as ex:
        if isinstance(ex, DefusedXmlException):
            url = 'https://pypi.python.org/pypi/defusedxml/0.3#attack-vectors'
            reason = 'Malicious XML was detected, see {0}.'.format(url)
            line = 0
        else:
            reason = ('Try validating your RDF with the W3 validator: '
                      'http://www.w3.org/RDF/Validator/.')
            line = ex.line()
        err.error(
                err_id=('main', 'test_package', 'parse_error'),
                error='Could not parse `install.rdf`.',
                description=('The RDF parser was unable to parse the '
                             'install.rdf file included with this add-on.',
                             reason),
                filename='install.rdf',
                line=line)
        return
    else:
        if install_rdf.rdf is None:
            err.error(
                    err_id=('main', 'test_package', 'cannot_parse_installrdf'),
                    error='Cannot read `install.rdf`',
                    description='The install.rdf file could not be parsed.',
                    filename='install.rdf')
            return
        else:
            err.save_resource('has_install_rdf', True, pushable=True)
            err.save_resource('install_rdf', install_rdf, pushable=True)

    # Load up the results of the type detection
    results = detect_type(err, install_rdf, package)
    if results is None:
        err.error(
                err_id=('main', 'test_package', 'undeterminable_type'),
                error='Unable to determine add-on type',
                description='The type detection algorithm could not determine '
                            'the type of the add-on.')
        return
    else:
        err.detected_type = results

    # Compare the results of the low-level type detection to
    # that of the expectation and the assumption.
    if expectation not in (PACKAGE_ANY, results):
        err.warning(
            err_id=('main', 'test_package', 'extension_type_mismatch'),
            warning='Extension Type Mismatch',
            description=("We detected that the add-on's type does not match "
                         'the expected type.',
                         'Type "%s" expected, found "%s"' %
                         (types[expectation], types[results])))


def _load_package_json(err, package, expectation):
    raw_package_json = package.read('package.json')
    try:
        package_json = json.loads(raw_package_json)
    except ValueError:
        err.error(
            err_id=('main', 'test_package', 'parse_error'),
            error='Could not parse `package.json`.',
            description='The JSON parser was unable to parse the '
                        'package.json file included with this add-on.',
            filename='package.json')
    else:
        err.save_resource('has_package_json', True, pushable=True)
        err.save_resource('package_json', package_json, pushable=True)
        err.detected_type = PACKAGE_EXTENSION


def _load_manifest_json(err, package, expectation):
    raw_manifest_json = package.read('manifest.json')
    try:
        manifest_json = ManifestJsonParser(err, raw_manifest_json)
    except ValueError:
        err.error(
            err_id=('main', 'test_package', 'parse_error'),
            error='Could not parse `manifest.json`.',
            description='The JSON parser was unable to parse the '
                        'manifest.json file included with this add-on.',
            filename='manifest.json')
    else:
        err.save_resource('has_manifest_json', True, pushable=True)
        err.save_resource('manifest_json', manifest_json, pushable=True)
        err.detected_type = PACKAGE_EXTENSION


def populate_chrome_manifest(err, xpi_package):
    "Loads the chrome.manifest if it's present"

    if 'chrome.manifest' in xpi_package:
        chrome_data = xpi_package.read('chrome.manifest')
        chrome = ChromeManifest(chrome_data, 'chrome.manifest')

        chrome_recursion_buster = set()

        # Handle the case of manifests linked from the manifest.
        def get_linked_manifest(path, from_path, from_chrome, from_triple):

            if path in chrome_recursion_buster:
                err.warning(
                    err_id=('submain', 'populate_chrome_manifest',
                            'recursion'),
                    warning='Linked manifest recursion detected.',
                    description='A chrome registration file links back to '
                                'itself. This can cause a multitude of '
                                'issues.',
                    filename=path)
                return

            # Make sure the manifest is properly linked
            if path not in xpi_package:
                err.notice(
                    err_id=('submain', 'populate_chrome_manifest', 'linkerr'),
                    notice='Linked manifest could not be found.',
                    description=('A linked manifest file could not be found '
                                 'in the package.',
                                 'Path: %s' % path),
                    filename=from_path,
                    line=from_triple['line'],
                    context=from_chrome.context)
                return

            chrome_recursion_buster.add(path)

            manifest = ChromeManifest(xpi_package.read(path), path)
            for triple in manifest.triples:
                yield triple

                if triple['subject'] == 'manifest':
                    subpath = triple['predicate']
                    # If the path is relative, make it relative to the current
                    # file.
                    if not subpath.startswith('/'):
                        subpath = '%s/%s' % (
                            '/'.join(path.split('/')[:-1]), subpath)

                    subpath = subpath.lstrip('/')

                    for subtriple in get_linked_manifest(
                            subpath, path, manifest, triple):
                        yield subtriple

            chrome_recursion_buster.discard(path)

        chrome_recursion_buster.add('chrome.manifest')

        # Search for linked manifests in the base manifest.
        for extra_manifest in chrome.get_triples(subject='manifest'):
            # When one is found, add its triples to our own.
            for triple in get_linked_manifest(extra_manifest['predicate'],
                                              'chrome.manifest', chrome,
                                              extra_manifest):
                chrome.triples.append(triple)

        chrome_recursion_buster.discard('chrome.manifest')

        # Create a reference so we can get the chrome manifest later, but make
        # it pushable so we don't run chrome manifests in JAR files.
        err.save_resource('chrome.manifest', chrome, pushable=True)
        # Create a non-pushable reference for tests that need to access the
        # chrome manifest from within JAR files.
        err.save_resource('chrome.manifest_nopush', chrome, pushable=False)


def test_inner_package(err, xpi_package, for_appversions=None):
    "Tests a package's inner content."

    populate_chrome_manifest(err, xpi_package)

    # Iterate through each tier.
    for tier in sorted(decorator.get_tiers()):

        # Let the error bundler know what tier we're on.
        err.set_tier(tier)

        # Iterate through each test of our detected type.
        for test in decorator.get_tests(tier, err.detected_type):
            # Test whether the test is app/version specific.
            if test['versions'] is not None:
                # If the test's version requirements don't apply to the add-on,
                # then skip the test.
                if not err.supports_version(test['versions']):
                    continue

                # If the user's version requirements don't apply to the test or
                # to the add-on, then skip the test.
                if (for_appversions and
                    not (err._compare_version(requirements=for_appversions,
                                              support=test['versions']) and
                         err.supports_version(for_appversions))):
                    continue

            # Save the version requirements to the error bundler.
            err.version_requirements = test['versions']

            test_func = test['test']
            if test['simple']:
                test_func(err)
            else:
                # Pass in:
                # - Error Bundler
                # - A copy of the package itself
                test_func(err, xpi_package)

        # Return any errors at the end of the tier if undetermined.
        if err.failed(fail_on_warnings=False) and not err.determined:
            err.unfinished = True
            err.discard_unused_messages(ending_tier=tier)
            return err

    # Return the results.
    return err



from functools import partial

from validator import decorator
from validator import metadata_helpers


@decorator.register_test(tier=1)
def test_manifest_json_params(err, xpi_manifest=None):
    if err.get_resource('has_manifest_json'):
        validate_required_id(err)
        validate_required_field(err, 'name', validate_name)
        validate_required_field(err, 'version', validate_version)


def validate_required_id(err):
    manifest_json = err.get_resource('manifest_json').data
    if ('applications' in manifest_json and
            'gecko' in manifest_json['applications'] and
            'id' in manifest_json['applications']['gecko']):
        value = manifest_json['applications']['gecko']['id']
        validate_id(err, value)
    else:
        create_missing_field_error(err, 'id')


def validate_required_field(err, field, validate_fn):
    manifest_json = err.get_resource('manifest_json').data
    if field in manifest_json:
        value = manifest_json[field]
        validate_fn(err, value)
    else:
        create_missing_field_error(err, field)


def create_missing_field_error(err, field):
    err.error(
        ('manifest_json', 'field_required', field),
        'Your manifest.json is missing a required field',
        'Your manifest.json is missing the "{field}" field.'.format(
            field=field))


validate_id = partial(metadata_helpers.validate_id, source='manifest.json')
validate_version = partial(
    metadata_helpers.validate_version, source='manifest.json')
validate_name = partial(metadata_helpers.validate_name, source='manifest.json')



import simplejson as json

from spidermonkey import Spidermonkey

from validator import unicodehelper
from validator.contextgenerator import ContextGenerator
from validator.decorator import register_cleanup


def get_tree(code, err=None, filename=None, shell=None):
    """Retrieve the parse tree for a JS snippet."""

    try:
        return JSShell.get_shell().get_tree(code)
    except JSReflectException as exc:
        str_exc = str(exc)
        if 'SyntaxError' in str_exc or 'ReferenceError' in str_exc:
            err.warning(('testcases_scripting', 'test_js_file',
                         'syntax_error'),
                        'JavaScript Compile-Time Error',
                        ['A compile-time error in the JavaScript halted '
                         'validation of that file.',
                         'Message: %s' % str_exc.split(':', 1)[-1].strip()],
                        filename=filename,
                        line=exc.line,
                        context=ContextGenerator(code))
        elif 'InternalError: too much recursion' in str_exc:
            err.notice(('testcases_scripting', 'test_js_file',
                        'recursion_error'),
                       'JS too deeply nested for validation',
                       'A JS file was encountered that could not be valiated '
                       'due to limitations with Spidermonkey. It should be '
                       'manually inspected.',
                       filename=filename)
        else:
            err.error(('testcases_scripting', 'test_js_file',
                       'retrieving_tree'),
                      'JS reflection error prevented validation',
                      ['An error in the JavaScript file prevented it from '
                       'being properly read by the Spidermonkey JS engine.',
                       str(exc)],
                      filename=filename)


@register_cleanup
class JSShell(Spidermonkey):
    instance = None

    SCRIPT = """
        function output(object) {
            print(JSON.stringify(object));
        }
        var stdin;
        while ((stdin = readline())) {
            try{
                stdin = JSON.parse(stdin);
                output(Reflect.parse(stdin));
            } catch(e) {
                output({
                    "error": true,
                    "error_message": e.toString(),
                    "line_number": e.lineNumber
                });
            }
        }
    """

    def __init__(self):
        # Use "version(180)" so we don't use the latest version (185 at the
        # time of this writing:
        # https://developer.mozilla.org/en-US/docs/Mozilla/Projects/SpiderMonkey/JSAPI_reference/JSVersion)  # noqa
        # which deprecates generators with 'function' (see
        # https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/Legacy_generator_function)  # noqa
        super(JSShell, self).__init__(code=['version(180)', self.SCRIPT])

    def __del__(self):
        if self.returncode is None:
            self.terminate()
        super(Spidermonkey, self).__del__()

    @classmethod
    def get_shell(cls):
        """Get a running JSShell instance, or create a new one if one does not
        already exist."""

        if not cls.instance:
            cls.instance = cls()

        return cls.instance

    @classmethod
    def cleanup(cls):
        """Clear our saved instance, and terminate its Spidermonkey process,
        if there are no further references to it."""
        cls.instance = None

    def get_tree(self, code):
        if isinstance(code, str):
            code = unicodehelper.decode(code)

        try:
            self.stdin.write(json.dumps(code))
            self.stdin.write('\n')

            output = json.loads(self.stdout.readline(), strict=False)
        except Exception:
            # If this instance is the cached instance, clear it.
            if self == self.__class__.instance:
                self.__class__.instance = None
            raise

        if output.get('error'):
            raise JSReflectException(output['error_message'],
                                     output['line_number'])

        return output


class JSReflectException(Exception):
    'An exception to indicate that tokenization has failed'

    def __init__(self, value, line=None):
        self.value = value
        self.line = line

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return '<JSReflectException (line {0}) {1!r}>'.format(
            self.line, self.value)



from helper import CompatTestCase
from validator.compat import TB31_DEFINITION


class TestTB31Compat(CompatTestCase):
    """Test that compatibility tests for Thunderbird 31 are properly executed."""

    VERSION = TB31_DEFINITION

    def test_nsIMsgCompose_checkAndPopulateRecipients(self):
        for method in self.run_xpcom_for_compat(
                'nsIAddrDatabase', ['addAllowRemoteContent']):
            self.assert_silent()
            self.assert_compat_error(type_='warning')

    def test_getNonHtmlRecipients(self):
        """Test that these patterns are flagged in Thunderbird 31."""
        self.run_script_for_compat('allowRemoteContentForSender()')
        self.assert_silent()
        self.assert_compat_error(type_='warning')

    def test_recentfolders_label(self):
        self.run_regex_for_compat('folderContextProperties.label')
        self.assert_silent()
        self.assert_compat_error(type_='warning')



import argparse
import json
import logging
import sys

from validator import constants
from validator.validate import validate


def main():
    'Main function. Handles delegation to other functions.'

    logging.basicConfig()

    type_choices = {'any': constants.PACKAGE_ANY,
                    'extension': constants.PACKAGE_EXTENSION,
                    'theme': constants.PACKAGE_THEME,
                    'dictionary': constants.PACKAGE_DICTIONARY,
                    'languagepack': constants.PACKAGE_LANGPACK,
                    'search': constants.PACKAGE_SEARCHPROV,
                    'multi': constants.PACKAGE_MULTI}

    # Parse the arguments that
    parser = argparse.ArgumentParser(
        description='Run tests on a Mozilla-type addon.')

    parser.add_argument('package',
                        help="The path of the package you're testing")
    parser.add_argument('-t',
                        '--type',
                        default='any',
                        choices=type_choices.keys(),
                        help="Type of addon you assume you're testing",
                        required=False)
    parser.add_argument('-o',
                        '--output',
                        default='text',
                        choices=('text', 'json'),
                        help='The output format that you expect',
                        required=False)
    parser.add_argument('-v',
                        '--verbose',
                        action='store_const',
                        const=True,
                        help="""If the output format supports it, makes
                        the analysis summary include extra info.""")
    parser.add_argument('--boring',
                        action='store_const',
                        const=True,
                        help="""Activating this flag will remove color
                        support from the terminal.""")
    parser.add_argument('--determined',
                        action='store_const',
                        const=True,
                        help="""This flag will continue running tests in
                        successive tests even if a lower tier fails.""")
    parser.add_argument('--selfhosted',
                        action='store_const',
                        const=True,
                        help="""Indicates that the addon will not be
                        hosted on addons.mozilla.org. This allows the
                        <em:updateURL> element to be set.""")
    parser.add_argument('--approved_applications',
                        help="""A JSON file containing acceptable applications
                        and their versions""",
                        required=False)
    parser.add_argument('--target-maxversion',
                        help="""JSON string to override the package's
                        targetapp_maxVersion for validation. The JSON object
                        should be a dict of versions keyed by application
                        GUID. For example, setting a package's max Firefox
                        version to 5.*:
                        {"{ec8030f7-c20a-464f-9b0e-13a3a9e97384}": "5.*"}
                        """)
    parser.add_argument('--target-minversion',
                        help="""JSON string to override the package's
                        targetapp_minVersion for validation. The JSON object
                        should be a dict of versions keyed by application
                        GUID. For example, setting a package's min Firefox
                        version to 5.*:
                        {"{ec8030f7-c20a-464f-9b0e-13a3a9e97384}": "5.*"}
                        """)
    parser.add_argument('--for-appversions',
                        help="""JSON string to run validation tests for
                        compatibility with a specific app/version. The JSON
                        object should be a dict of version lists keyed by
                        application GUID. For example, running Firefox 6.*
                        compatibility tests:
                        {"{ec8030f7-c20a-464f-9b0e-13a3a9e97384}": ["6.*"]}
                        """)
    parser.add_argument('--timeout',
                        help='The amount of time before validation is '
                             'terminated with a timeout exception.',
                        default='60')

    args = parser.parse_args()

    # We want to make sure that the output is expected. Parse out the expected
    # type for the add-on and pass it in for validation.
    if args.type not in type_choices:
        # Fail if the user provided invalid input.
        print 'Given expectation (%s) not valid. See --help for details' % \
                args.type
        sys.exit(1)

    overrides = {}
    if args.target_minversion:
        overrides['targetapp_minVersion'] = json.loads(args.target_minversion)
    if args.target_maxversion:
        overrides['targetapp_maxVersion'] = json.loads(args.target_maxversion)

    for_appversions = None
    if args.for_appversions:
        for_appversions = json.loads(args.for_appversions)

    try:
        timeout = int(args.timeout)
    except ValueError:
        print 'Invalid timeout. Integer expected.'
        sys.exit(1)

    expectation = type_choices[args.type]
    error_bundle = validate(args.package,
                            format=None,
                            approved_applications=args.approved_applications,
                            determined=args.determined,
                            listed=not args.selfhosted,
                            overrides=overrides,
                            for_appversions=for_appversions,
                            expectation=expectation,
                            timeout=timeout)

    # Print the output of the tests based on the requested format.
    if args.output == 'text':
        print error_bundle.print_summary(verbose=args.verbose,
                                         no_color=args.boring).encode('utf-8')
    elif args.output == 'json':
        sys.stdout.write(error_bundle.render_json())

    if error_bundle.failed():
        sys.exit(1)
    else:
        sys.exit(0)

# Start up the testing and return the output.
if __name__ == '__main__':
    main()



from validator.decorator import version_range
from validator.constants import (FIREFOX_GUID, FENNEC_GUID,
                                 THUNDERBIRD_GUID as TB_GUID, ANDROID_GUID)


# Compatibility app/version ranges:

def _build_definition(maj_version_num, firefox=True, fennec=True,
                      thunderbird=True, android=True):
    definition = {}
    app_version_range = (
        lambda app: version_range(app, '%d.0a1' % maj_version_num,
                                       '%d.0a1' % (maj_version_num + 1)))
    if firefox:
        definition[FIREFOX_GUID] = app_version_range('firefox')
    if fennec:
        definition[FENNEC_GUID] = app_version_range('fennec')
    if thunderbird:
        definition[TB_GUID] = app_version_range('thunderbird')
    if android:
        definition[ANDROID_GUID] = app_version_range('android')

    return definition


FX38_DEFINITION = _build_definition(38)
FX39_DEFINITION = _build_definition(39)
FX40_DEFINITION = _build_definition(40)
FX41_DEFINITION = _build_definition(41)
FX42_DEFINITION = _build_definition(42)
FX43_DEFINITION = _build_definition(43)
FX44_DEFINITION = _build_definition(44)
FX45_DEFINITION = _build_definition(45)
FX46_DEFINITION = _build_definition(46)


_tb_definition = (lambda ver:
    _build_definition(ver, firefox=False, fennec=False, android=False))

TB29_DEFINITION = _tb_definition(29)
TB30_DEFINITION = _tb_definition(30)
TB31_DEFINITION = _tb_definition(31)



from xml.parsers.expat import ExpatError

from defusedxml.minidom import parse
from defusedxml.common import DefusedXmlException

from validator.constants import *


def detect_opensearch(err, package, listed=False):
    'Detect, parse, and validate an OpenSearch provider'

    # Parse the file.
    try:
        # Check if it is a file object.
        if hasattr(package, 'read'):
            srch_prov = parse(package)
        else:
            # It's not a file object; open it (the XML parser is bad at this).
            with open(package, 'rb') as package_file:
                srch_prov = parse(package_file)
    except DefusedXmlException:
        url = 'https://pypi.python.org/pypi/defusedxml/0.3#attack-vectors'
        err.error(
            err_id=('opensearch', 'security_error'),
            error='OpenSearch: XML Security Error',
            description='The OpenSearch extension could not be parsed due to '
                        'a security error in the XML. See {url} for more '
                        'info.'.format(url=url))
        return err
    except ExpatError:
        err.error(
            err_id=('opensearch', 'parse_error'),
            error='OpenSearch: XML Parse Error',
            description='The OpenSearch extension could not be parsed due to '
                        'a syntax error in the XML.')
        return err

    # Make sure that the root element is OpenSearchDescription.
    if srch_prov.documentElement.tagName != 'OpenSearchDescription':
        err.error(
            err_id=('opensearch', 'invalid_document_root'),
            error='OpenSearch: Invalid Document Root',
            description='The root element of the OpenSearch provider is not '
                        "'OpenSearchDescription'.")

    # Per bug 617822
    if not srch_prov.documentElement.hasAttribute('xmlns'):
        err.error(
            err_id=('opensearch', 'no_xmlns'),
            error='OpenSearch: Missing XMLNS attribute',
            description='The XML namespace attribute is missing from the '
                        'OpenSearch document.')

    if ('xmlns' not in srch_prov.documentElement.attributes.keys() or
        srch_prov.documentElement.attributes['xmlns'].value not in (
                    'http://a9.com/-/spec/opensearch/1.0/',
                    'http://a9.com/-/spec/opensearch/1.1/',
                    'http://a9.com/-/spec/opensearchdescription/1.1/',
                    'http://a9.com/-/spec/opensearchdescription/1.0/')):
        err.error(
            err_id=('opensearch', 'invalid_xmlns'),
            error='OpenSearch: Bad XMLNS attribute',
            description='The XML namespace attribute contains an '
                        'value.')

    # Make sure that there is exactly one ShortName.
    sn = srch_prov.documentElement.getElementsByTagName('ShortName')
    if not sn:
        err.error(
            err_id=('opensearch', 'missing_shortname'),
            error='OpenSearch: Missing <ShortName> elements',
            description='ShortName elements are mandatory OpenSearch provider '
                        'elements.')
    elif len(sn) > 1:
        err.error(
            err_id=('opensearch', 'extra_shortnames'),
            error='OpenSearch: Too many <ShortName> elements',
            description='Too many ShortName elements exist in the OpenSearch '
                        'provider.')
    else:
        sn_children = sn[0].childNodes
        short_name = 0
        for node in sn_children:
            if node.nodeType == node.TEXT_NODE:
                short_name += len(node.data)
        if short_name > 16:
            err.error(
                err_id=('opensearch', 'big_shortname'),
                error='OpenSearch: <ShortName> element too long',
                description='The ShortName element must contains less than '
                            'seventeen characters.')

    # Make sure that there is exactly one Description.
    if len(srch_prov.documentElement.getElementsByTagName('Description')) != 1:
        err.error(
            err_id=('opensearch', 'missing_description'),
            error='OpenSearch: Invalid number of <Description> elements',
            description='There are too many or too few Description elements '
                        'in the OpenSearch provider.')

    # Grab the URLs and make sure that there is at least one.
    urls = srch_prov.documentElement.getElementsByTagName('Url')
    if not urls:
        err.error(
            err_id=('opensearch', 'missing_url'),
            error='OpenSearch: Missing <Url> elements',
            description='The OpenSearch provider is missing a Url element.')

    if listed and any(url.hasAttribute('rel') and
                      url.attributes['rel'].value == 'self' for
                      url in urls):
        err.error(
            err_id=('opensearch', 'rel_self'),
            error='OpenSearch: <Url> elements may not be rel=self',
            description='Per AMO guidelines, OpenSearch providers cannot '
                        "contain <Url /> elements with a 'rel' attribute "
                        "pointing to the URL's current location. It must be "
                        'removed before posting this provider to AMO.')

    acceptable_mimes = ('text/html', 'application/xhtml+xml')
    acceptable_urls = [url for url in urls if url.hasAttribute('type') and
                          url.attributes['type'].value in acceptable_mimes]

    # At least one Url must be text/html
    if not acceptable_urls:
        err.error(
            err_id=('opensearch', 'missing_url_texthtml'),
            error="OpenSearch: Missing <Url> element with 'text/html' type",
            description='OpenSearch providers must have at least one Url '
                        "element with a type attribute set to 'text/html'.")

    # Make sure that each Url has the require attributes.
    for url in acceptable_urls:

        if url.hasAttribute('rel') and url.attributes['rel'].value == 'self':
            continue

        if url.hasAttribute('method') and \
           url.attributes['method'].value.upper() not in ('GET', 'POST'):
            err.error(
                err_id=('opensearch', 'missing_method'),
                error="OpenSearch: <Url> element with invalid 'method'",
                description='A Url element in the OpenSearch provider lists a '
                            'method attribute, but the value is not GET or '
                            'POST.')

        # Test for attribute presence.
        if not url.hasAttribute('template'):
            err.error(
                err_id=('opensearch', 'missing_template'),
                error='OpenSearch: <Url> element missing template attribute',
                description='<Url> elements of OpenSearch providers must '
                            'include a template attribute.')
        else:
            url_template = url.attributes['template'].value
            if url_template[:4] != 'http':
                err.error(
                    err_id=('opensearch', 'invalid_template'),
                    error='OpenSearch: `<Url>` element with invalid '
                          '`template`',
                    description='A `<Url>` element in the OpenSearch '
                                'provider lists a template attribute, but '
                                'the value is not a valid HTTP URL.')

            # Make sure that there is a {searchTerms} placeholder in the
            # URL template.
            found_template = url_template.count('{searchTerms}') > 0

            # If we didn't find it in a simple parse of the template=""
            # attribute, look deeper at the <Param /> elements.
            if not found_template:
                for param in url.getElementsByTagName('Param'):
                    # As long as we're in here and dependent on the
                    # attributes, we'd might as well validate them.
                    attribute_keys = param.attributes.keys()
                    if 'name' not in attribute_keys or \
                       'value' not in attribute_keys:
                        err.error(
                            err_id=('opensearch', 'param_missing_attrs'),
                            error='OpenSearch: `<Param>` element missing '
                                  'name/value',
                            description='Param elements in the OpenSearch '
                                        'provider must include a name and a '
                                        'value attribute.')

                    param_value = (param.attributes['value'].value if
                                   'value' in param.attributes.keys() else
                                   '')
                    if param_value.count('{searchTerms}'):
                        found_template = True

                        # Since we're in a validating spirit, continue
                        # looking for more errors and don't break

            # If the template still hasn't been found...
            if not found_template:
                tpl = url.attributes['template'].value
                err.error(
                    err_id=('opensearch', 'template_not_found'),
                    error='OpenSearch: <Url> element missing template '
                          'placeholder',
                    description=('`<Url>` elements of OpenSearch providers '
                                 'must include a template attribute or '
                                 'specify a placeholder with '
                                 '`{searchTerms}`.',
                                 'Missing template: %s' % tpl))

    # Make sure there are no updateURL elements
    if srch_prov.getElementsByTagName('updateURL'):
        err.error(
            err_id=('opensearch', 'banned_updateurl'),
            error='OpenSearch: <updateURL> elements are banned in OpenSearch '
                  'providers.',
            description='OpenSearch providers may not contain <updateURL> '
                        'elements.')

    # The OpenSearch provider is valid!
    return err




from helper import CompatTestCase
from validator.compat import FX44_DEFINITION


class TestFX44Compat(CompatTestCase):
    """Test that compatibility tests for Gecko 44 are properly executed."""

    VERSION = FX44_DEFINITION

    def test_new_devtools_location_gre(self):
        self.run_script_for_compat("""
            var devtools = Cu.import("resource://gre/modules/devtools/shared/Loader.jsm", {}).devtools;
        """)
        self.assert_silent()
        self.assert_compat_error()

    def test_new_devtools_location_from_root(self):
        self.run_script_for_compat("""
            var devtools = Cu.import("resource:///modules/devtools/Target.jsm", {});
        """)
        self.assert_silent()
        self.assert_compat_error()

    def test_noSuchMethod_literal(self):
        self.run_script_for_compat("""
            var o = {
              __noSuchMethod__: function(id, args) {
                console.log(id, '(' + args.join(', ') + ')');
              }
          };
        """)
        self.assert_silent()
        self.assert_compat_error()

    def test_noSuchMethod_assignment(self):
        self.run_script_for_compat("""
            var o = {};
            o.__noSuchMethod__ = function(id, args) {
              console.log(id, '(' + args.join(', ') + ')');
            };
        """)
        self.assert_silent()
        self.assert_compat_error()

    def test_noSuchMethod_string_assignment(self):
        self.run_script_for_compat("""
            var o = {};
            o['__noSuchMethod__'] = function(id, args) {
              console.log(id, '(' + args.join(', ') + ')');
            };
        """)

    def test_getAllStyleSheets(self):
        self.run_script_for_compat(
            'let stylesheets = window.getAllStyleSheets(config.browser.contentWindow);'
        )
        self.assert_silent()
        self.assert_compat_error()



import json

from validator.constants import FIREFOX_GUID


class ManifestJsonParser(object):
    """Parser wrapper for manifest.json files."""

    def __init__(self, err, data, namespace=None):
        self.err = err
        self.data = json.loads(data)

    def get_applications(self):
        """Return the list of supported applications."""
        if ('applications' not in self.data or
                'gecko' not in self.data['applications']):
            return []
        app = self.data['applications']['gecko']
        min_version = app.get('strict_min_version', u'42.0')
        max_version = app.get('strict_max_version', u'*')
        return [{u'guid': FIREFOX_GUID,
                 u'min_version': min_version,
                 u'max_version': max_version}]



from call_definitions import open_in_chrome_context
from instanceproperties import _set_HTML_property
from validator.compat import (FX42_DEFINITION, TB29_DEFINITION,
                              TB30_DEFINITION, TB31_DEFINITION)
from validator.constants import BUGZILLA_BUG


ENTITIES = {}


def register_entity(name):
    """Allow an entity's modifier to be registered for use."""
    def wrap(func):
        ENTITIES[name] = func
        return func
    return wrap


def entity(name, result=None):
    assert name in ENTITIES

    def return_wrap(t):
        output = ENTITIES[name](traverser=t)
        if result is not None:
            return result
        elif output is not None:
            return output
        else:
            return {'value': {}}
    return {'value': return_wrap}


def deprecated_entity(name, version, message, bug, status='deprecated',
                      compat_type='error'):
    def wrap(traverser):
        traverser.err.warning(
            err_id=('js', 'entities', name),
            warning='`%s` has been %s.' % (name, status),
            description=(message,
                         'See %s for more information.' % BUGZILLA_BUG % bug),
            filename=traverser.filename,
            line=traverser.line,
            column=traverser.position,
            context=traverser.context,
            for_appversions=version,
            compatibility_type=compat_type,
            tier=5)
    register_entity(name)(wrap)

def register_changed_entities(version_definition, entities, version_string):
    for entity in entities:
        deprecated_entity(
            name=entity['name'],
            version=version_definition,
            message='The method or property `%s` has been `%s` in `%s`.'
                % (entity['name'], entity['status'], version_string),
            bug=entity['bug'],
            compat_type=entity['compat_type'])


DOC_WRITE_MSG = ('https://developer.mozilla.org/docs/XUL/School_tutorial/'
                 'DOM_Building_and_HTML_Insertion')

@register_entity('document.write')
def document_write(traverser):
    def on_write(wrapper, arguments, traverser):
        traverser.err.warning(
            err_id=('js', 'document.write', 'evil'),
            warning='Use of `document.write` strongly discouraged.',
            description=('`document.write` will fail in many circumstances ',
                         'when used in extensions, and has potentially severe '
                         'security repercussions when used improperly. '
                         'Therefore, it should not be used. See %s for more '
                         'information.' % DOC_WRITE_MSG),
            filename=traverser.filename,
            line=traverser.line,
            column=traverser.position,
            context=traverser.context)
        if not arguments:
            return
        value = traverser._traverse_node(arguments[0])
        _set_HTML_property('document.write()', value, traverser)

    return {'return': on_write}


@register_entity('nsIDNSService.resolve')
def nsIDNSServiceResolve(traverser):
    traverser.err.warning(
        err_id=('testcases_javascript_entity_values',
                'nsIDNSServiceResolve'),
        warning='`nsIDNSService.resolve()` should not be used.',
        description='The `nsIDNSService.resolve` method performs a '
                    'synchronous DNS lookup, which will freeze the UI. This '
                    'can result in severe performance issues. '
                    '`nsIDNSService.asyncResolve()` should be used instead.',
        filename=traverser.filename,
        line=traverser.line,
        column=traverser.position,
        context=traverser.context)


@register_entity('nsISound.play')
def nsISoundPlay(traverser):
    traverser.err.warning(
        err_id=('testcases_javascript_entity_values',
                'nsISound_play'),
        warning='`nsISound.play` should not be used.',
        description='The `nsISound.play` function is synchronous, and thus '
                    'freezes the interface while the sound is playing. It '
                    'should be avoided in favor of the HTML5 audio APIs.',
        filename=traverser.filename,
        line=traverser.line,
        column=traverser.position,
        context=traverser.context)


@register_entity('nsIWindowWatcher.openWindow')
def nsIWindowWatcher_openWindow(traverser):
    def on_open(wrapper, arguments, traverser):
        if not arguments:
            return
        uri = traverser._traverse_node(arguments[0])
        open_in_chrome_context(uri, 'nsIWindowWatcher.openWindow', traverser)

    return {'return': on_open}


@register_entity('nsITransferable.init')
def nsITransferable_init(traverser):
    def on_init(wrapper, arguments, traverser):
        if not arguments:
            return
        first_arg = traverser._traverse_node(arguments[0])
        if first_arg.get_literal_value():
            return
        traverser.err.warning(
            err_id=('js_entity_values', 'nsITransferable', 'init'),
            warning='`init` should not be called with a null first argument',
            description='Calling `nsITransferable.init()` with a null first '
                        'argument has the potential to leak data across '
                        'private browsing mode sessions. `null` is  '
                        'appropriate only when reading data or writing data '
                        'which is not associated with a particular window.',
            filename=traverser.filename,
            line=traverser.line,
            column=traverser.position,
            context=traverser.context)

    return {'return': on_init}


@register_entity('NewTabURL.override')
def NewTabURL_override(traverser):
    def on_override(wrapper, arguments, traverser):
        # Import loop.
        from validator.testcases.javascript.predefinedentities import (
            CUSTOMIZATION_API_HELP)
        traverser.err.warning(
            err_id=('js_entity_values', 'NewTabURL', 'override'),
            warning='Extensions must not alter user preferences such as the '
                    'new tab URL without explicit user consent.',
            description='Extensions must not alter user preferences such as '
                        'the new tab URL without explicit user consent. Such '
                        'changes must also be reverted when the extension is '
                        'disabled or uninstalled.',
            signing_severity='high',
            signing_help='Add-ons which directly change these preferences must '
                         'undergo manual code review for at least one '
                         'submission. ' + CUSTOMIZATION_API_HELP,
        )
    return {'return': on_override}


@register_entity('nsIObserverService.addObserver')
def nsIObserverService_addObserver(traverser):
    def on_addObserver(wrapper, arguments, traverser):
        if not arguments:
            return

        first_arg = traverser._traverse_node(arguments[1]).get_literal_value()

        if first_arg == 'newtab-url-changed':
            traverser.err.warning(
                err_id=('js_entity_values', 'nsIObserverService', 'newtab_url_changed'),
                warning='Extensions must not use the `newtab-url-changed` event to '
                        'revert changes made to the new tab url.',
                description='To avoid conflicts, extensions are not allowed to '
                            'add an observer to the `newtab-url-changed` event '
                            'in order to revert changes that have been made to '
                            'the new tab url.',
                signing_severity='high',
                signing_help='Add-ons which use `newtab-url-changed` to change '
                             'the new tab url are not allowed.')

    # Return the addObserver handler and a general dangerous warning.
    return {
        'return': on_addObserver,
        'dangerous': lambda a, t, e:
            e.get_resource('em:bootstrap') and
            'Authors of bootstrapped add-ons must take care '
            'to remove any added observers '
            'at shutdown.'
    }


@register_entity('nsIPermissionManager.add')
@register_entity('nsIPermissionManager.addFromPrincipal')
@register_entity('nsIPermissionManager.remove')
@register_entity('nsIPermissionManager.removeFromPrincipal')
@register_entity('nsIPermissionManager.removeAll')
@register_entity('nsIPermissionManager.testExactPermission')
@register_entity('nsIPermissionManager.testExactPermissionFromPrincipal')
@register_entity('nsIPermissionManager.testPermission')
@register_entity('nsIPermissionManager.testPermissionFromPrincipal')
def nsIPermissionManager(traverser):
    traverser.err.warning(
        err_id=('testcases_javascript_entity_values',
                'nsIPermissionManager'),
        warning='The Permission Manager now uses origins instead of hosts.',
        description=(
            'The Permission Manager now uses origins instead of hosts, '
            'meaning that http://mozilla.com and https://mozilla.com now '
            'have different permissions. Some methods in the component now '
            'take nsIURI instead of strings. '
            'See %s for more information.' % BUGZILLA_BUG % 1165263),
        filename=traverser.filename,
        line=traverser.line,
        column=traverser.position,
        for_appversions=FX42_DEFINITION,
        compatibility_type='error',
        context=traverser.context,
        tier=5)


# Thunderbird 29 JS changes
TB29_JS_ENTITIES = [
    {'name':'DisablePhishingWarning',
     'status': '',
     'bug': 926473,
     'compat_type': 'error'},
    {'name':'RoomInfo',
     'status': '',
     'bug': 920801,
     'compat_type': 'error'},
    {'name':'FillInHTMLTooltip',
     'status': '',
     'bug': 956586,
     'compat_type': 'error'},
    {'name':'escapeXMLchars',
     'status': '',
     'bug': 942638,
     'compat_type': 'error'},
    {'name':'gPluginHandler.isTooSmall',
     'status': 'removed',
     'bug': 951800,
     'compat_type': 'error'},
    {'name':'XMPPSession.authDialog',
     'status': 'removed',
     'bug': 920801,
     'compat_type': 'error'},
    {'name':'XMPPMUCConversation.supportChatStateNotifications',
     'status': 'removed',
     'bug': 920801,
     'compat_type': 'error'},
    {'name':'Socket.inputSegmentSize',
     'status': 'removed',
     'bug': 920801,
     'compat_type': 'error'},
    {'name':'XMPPMUCConversationPrototype.normalizedName',
     'status': 'removed',
     'bug': 957918,
     'compat_type': 'error'},
    {'name':'XMPPAccountBuddyPrototype.normalizedName',
     'status': 'removed',
     'bug': 957918,
     'compat_type': 'error'},
    {'name':'XMPPAccountPrototype.normalizedName',
     'status': 'removed',
     'bug': 957918,
     'compat_type': 'error'},
    {'name':'GenericAccountPrototype.maxMessageLength',
     'status': 'removed',
     'bug': 954484,
     'compat_type': 'error'},
    {'name':'mailTabType.desiredColumnStates',
     'status': 'removed',
     'bug': 528044,
     'compat_type': 'error'},
]
register_changed_entities(version_definition=TB29_DEFINITION,
    entities=TB29_JS_ENTITIES, version_string='Thunderbird 29')

# Thunderbird 30 IDL changes
TB30_ENTITIES = [
    {'name':'nsIMsgDatabase.forceFolderDBClosed',
     'status': 'moved to nsIMsgDBService',
     'bug': 876548,
     'compat_type': 'error'},
    {'name':'nsIMsgCompose.checkAndPopulateRecipients',
     'status': 'removed',
     'bug': 970118,
     'compat_type': 'error'},
]
register_changed_entities(version_definition=TB30_DEFINITION,
    entities=TB30_ENTITIES, version_string='Thunderbird 30')

# Thunderbird 30 JS changes
TB30_JS_ENTITIES = [
    {'name':'GlodaMsgSearcher.retrievalLimit',
     'status': 'removed',
     'bug': 742236,
     'compat_type': 'error'},
    {'name':'GlodaIMSearcher.retrievalLimit',
     'status': 'removed',
     'bug': 742236,
     'compat_type': 'error'},
    {'name':'getNonHtmlRecipients',
     'status': 'removed',
     'bug': 970118,
     'compat_type': 'error'},
]
register_changed_entities(version_definition=TB30_DEFINITION,
    entities=TB30_JS_ENTITIES, version_string='Thunderbird 30')

# Thunderbird 31 IDL changes
TB31_ENTITIES = [
    {'name':'nsIFolderLookupService.getFolderById',
         'status': 'removed',
         'bug': 441437,
         'compat_type': 'error'},
    {'name':'nsIAbCard.kAllowRemoteContentProperty',
         'status': 'removed',
         'bug': 457296,
         'compat_type': 'error'},
    {'name':'nsIAddrDatabase.addAllowRemoteContent',
         'status': 'removed',
         'bug': 457296,
         'compat_type': 'error'},
]
register_changed_entities(version_definition=TB31_DEFINITION,
    entities=TB31_ENTITIES, version_string='Thunderbird 31')

# Thunderbird 31 JS changes
TB31_JS_ENTITIES = [
    {'name':'ircAccount.kFields.away',
         'status': 'removed',
         'bug': 955698,
         'compat_type': 'error'},
    {'name':'ircAccount.kFields.idleTime',
         'status': 'removed',
         'bug': 987577,
         'compat_type': 'error'},
    {'name':'updateCharsetPopupMenu',
         'status': 'renamed to EditorUpdateCharsetMenu',
         'bug': 992643,
         'compat_type': 'error'},
    {'name':'EditorSetDocumentCharacterSet',
         'status': 'renamed to EditorSetCharacterSet',
         'bug': 992643,
         'compat_type': 'error'},
    {'name':'EmailConfigWizard.onChangedAuth',
         'status': 'renamed to onChangedInAuth',
         'bug': 883670,
         'compat_type': 'error'},
    {'name':'EmailConfigWizard.onInputUsername',
         'status': 'renamed to onInputOutUsername',
         'bug': 883670,
         'compat_type': 'error'},
    {'name':'awArrowHit',
         'status': 'removed',
         'bug': 959209,
         'compat_type': 'error'},
    {'name':'UpdateMailEditCharset',
         'status': 'removed',
         'bug': 999881,
         'compat_type': 'error'},
    {'name':'InitCharsetMenuCheckMark',
         'status': 'removed',
         'bug': 999881,
         'compat_type': 'error'},
    {'name':'gSecurityPane.readAcceptCookies',
         'status': 'moved to gPrivacyPane.readAcceptCookies',
         'bug': 953426,
         'compat_type': 'error'},
    {'name':'gSecurityPane.writeAcceptCookies',
         'status': 'moved to gPrivacyPane.writeAcceptCookies',
         'bug': 953426,
         'compat_type': 'error'},
    {'name':'gSecurityPane.showCookieExceptions',
         'status': 'moved to gPrivacyPane.showCookieExceptions',
         'bug': 953426,
         'compat_type': 'error'},
    {'name':'gSecurityPane.showCookies',
         'status': 'moved to gPrivacyPane.showCookies',
         'bug': 953426,
         'compat_type': 'error'},
    {'name':'allowRemoteContentForSender',
         'status': 'removed',
         'bug': 457296,
         'compat_type': 'error'},
    {'name':'allowRemoteContentForSite',
         'status': 'renamed to allowRemoteContentForURI',
         'bug': 457296,
         'compat_type': 'error'},
    {'name':'createNewHeaderView',
         'status': 'renamed to HeaderView',
         'bug': 898860,
         'compat_type': 'error'},
    {'name':'FolderDisplayWidget.getVisibleRowPadding',
         'status': 'changed',
         'bug': 964824,
         'compat_type': 'error'},
]
register_changed_entities(version_definition=TB31_DEFINITION,
    entities=TB31_JS_ENTITIES, version_string='Thunderbird 31')



from tests.js_helper import TestCase


class TestJSArrays(TestCase):
    """Test that arrays in the JS engine behave properly."""

    def test_object_overloading(self):
        self.run_script("""
        var x = [];
        x[1] = "asdf";
        x["asdf"] = "zxcv";
        var a = x[1],
            b = x.asdf;
        """)
        self.assert_var_eq('a', 'asdf')
        self.assert_var_eq('b', 'zxcv')

    def test_stringification(self):
        self.run_script("""
        var x = [4];
        var a = x * 3;
        """)
        self.assert_var_eq('a', 12)
